//==========================================================================================
//  Chord Level Select for MuseScore 3.2.x or higher
//  Portions of this script were adapted from the colornotes.qml, walk.qml and
//  other scripts by Werner Schweer, et al + Rob Birdwell/BirdwellMusic
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License version 2
//  as published by the Free Software Foundation and appearing in
//  the file LICENCE.GPL
//===========================================================================================
import QtQuick 2.1
import QtQuick.Controls 1.0
import QtQuick.Dialogs 1.1
import MuseScore 3.0

MuseScore {
    version: "1.1"
    description: "Select [top/bottom] or any given level between [2-7] of notes from a chord based on their vertical stack (counting from bottom) in chord(s)."
    menuPath: "Plugins.Select Chord Levels"
    pluginType: "dialog"
    width:  500
    height: 215
    id: level

    Component.onCompleted: {
        if (mscoreMajorVersion >= 4) {
            level.title = qsTr("Select Chord Levels")
            //level.categoryCode = "composing-arranging-tools"
        }
    }

    function getAllChordsInRange(chordArray) {

        var cursor = curScore.newCursor();
        cursor.rewind(1);
        var startStaff;
        var endStaff;
        var endTick;

        if (!cursor.segment) { // no selection
            console.log("No valid single-staff region selected.");
            return;
        }

        startStaff = cursor.staffIdx;
        cursor.rewind(2);
        if (cursor.tick == 0) {
            // this happens when the selection includes
            // the last measure of the score.
            // rewind(2) goes behind the last segment (where
            // there's none) and sets tick=0
            endTick = curScore.lastSegment.tick + 1;
        } else {
            endTick = cursor.tick;
        }

        endStaff = cursor.staffIdx;

        if (startStaff != endStaff) {
            console.log("Required: Only one staff within selection range.");
            return;
        }

        for (var staff = startStaff; staff <= endStaff; staff++) {
            for (var voice = 0; voice < 4; voice++) {
                cursor.rewind(1); // sets voice to 0
                cursor.voice = voice; //voice has to be set after goTo
                cursor.staffIdx = staff;

                while (cursor.segment && (cursor.tick < endTick)) {
                    if (cursor.element && cursor.element.type == Element.CHORD) {
                        var graceChords = cursor.element.graceNotes;
                        for (var i = 0; i < graceChords.length; i++) {
                            chordArray.push(graceChords[i]);
                        }

                        // the chord of the notes...
                        chordArray.push(cursor.element);

                    }
                    cursor.next();
                }
            }
        }
    }


    function displayMessageDlg(msg) {
        ctrlMessageDialog.text = qsTr(msg);
        ctrlMessageDialog.visible = true;
    }

    function selectLevels(move, deleteDesired) {
        console.log("Starting selectLevels()");
        var operationPerformed = false;

        curScore.startCmd(); // Start collecting undo info.

        var chords = new Array();
        getAllChordsInRange(chords);
        if (!chords.length) {
            displayMessageDlg(qsTr("No valid range selection on current score! Tsk Tsk."));
            return;
        }

        var levels = new Array();
        if (ctrlCheckBoxLevel7.checked) {
            levels.push(7);
        }
        if (ctrlCheckBoxLevel6.checked) {
            levels.push(6);
        }
        if (ctrlCheckBoxLevel5.checked) {
            levels.push(5);
        }
        if (ctrlCheckBoxLevel4.checked) {
            levels.push(4);
        }
        if (ctrlCheckBoxLevel3.checked) {
            levels.push(3);
        }
        if (ctrlCheckBoxLevel2.checked) {
            levels.push(2);
        }
        if (ctrlCheckBoxLevel1.checked) {
            levels.push(1);
        }

        // Require a level to be selected when only selecting,
        if (!levels.length && !move && !ctrlCheckBoxTopLevel.checked) { // Range selection can be changed to given voice. It is redundant to built in commands, but why not? If attempting merely to select, however...
            displayMessageDlg(qsTr("No levels(s) checked! Select the level(s) that match your chord stack sizes."));
            return;
        }


        // Method: will clear the range selection and begin adding one at a time the notes that correlate with user-checked levels
        curScore.selection.clear();


        var emptyChordPotential = false;
        for (var c = 0; c < chords.length; c++) {
            var notesInChord = chords[c].notes.length;
            console.log("# of notes in chord # " + (c + 1) + ":" + notesInChord);
            var notesQueuedToDeleteInChord = 0;
            for (var n = 0; n < chords[c].notes.length; n++) {
                for (var j = 0; j < levels.length; j++) {
                    // skip if not in levels
                    if ((levels[j] - 1) != n)
                        continue;
                    // add note to list selection or delete it
                    else {
                        if (deleteDesired)
                            chords[c].remove(chords[c].notes[n]);
                        else curScore.selection.select(chords[c].notes[n], true);
                    }
                }
            }
            // Special handling for absolute top note [notes.length-1] ought to be the index of the top-most note of any given chord-level
            if (ctrlCheckBoxTopLevel.checked) { // Top is checked, add it to list
                if (deleteDesired)
                    chords[c].remove(chords[c].notes[chords[c].notes.length - 1]);
                else curScore.selection.select(chords[c].notes[chords[c].notes.length - 1], true);
            }
        }
        // Switch voice of remaining selection if Revoice
        if (move) {
            var cmdVoiceIndex = ctrlComboBoxVoice.currentIndex + 1;
            console.log("selectLevels() is attempting to move selected notes (in selected levels) to layer " + cmdVoiceIndex);
            cmd("voice-" + cmdVoiceIndex);
            console.log("selectLevels() cmd call was executed.");
        }

        curScore.endCmd(); // Finish off the undo record.

        operationPerformed = true;
        console.log("Ending selectLevels()");
        return operationPerformed;
    }


    onRun: {
        console.log("Chord Levels script starting...");

        if (typeof curScore === 'undefined') {
            var msg = "Chord Levels exiting without processing - no current score!";
            console.log(msg);
            displayMessageDlg(msg);
            quit();
        }
    }

    Rectangle {
        property alias mouseArea: mouseArea
        property alias btnDeleteLevels: btnDeleteLevels
        property alias btnRevoiceLevels: btnRevoiceLevels
        property alias btnClose: btnClose
        property alias ctrlHintLabel: ctrlHintLabel
        property alias ctrlMessageDialog: ctrlMessageDialog

        width: 600 // added 200 (from 400)
        height: 240
        color: "grey"

        MessageDialog {
            id: ctrlMessageDialog
            icon: StandardIcon.Information
            title: "Chord Levels Message"
            text: "Welcome to Chord Levels!"
            visible: false
            onAccepted: {
                visible = false;
            }
        }

        MouseArea {
            id: mouseArea
            anchors.rightMargin: 0
            anchors.bottomMargin: 0
            anchors.leftMargin: 0
            anchors.topMargin: 0
            anchors.fill: parent

            Text {
                id: ctrlStackRangeLabel
                x: 25
                y: 15
                width: 100
                text: "Levels:"
            }

            Column {
                x: 80
                y: 15
                CheckBox {
                    id: ctrlCheckBoxLevel1
                    text: qsTr("Bottom") // 1
                }
                CheckBox {
                    id: ctrlCheckBoxLevel4
                    text: qsTr("4")
                }
                CheckBox {
                    id: ctrlCheckBoxLevel3
                    text: qsTr("3")
                }
                CheckBox {
                    id: ctrlCheckBoxLevel2
                    text: qsTr("2")
                }
            }

            Column {
                x: 175 // origin 125
                y: 15
                CheckBox {
                    id: ctrlCheckBoxTopLevel
                    text: qsTr("Top") // 8
                }
                CheckBox {
                    id: ctrlCheckBoxLevel7
                    text: qsTr("7")
                }
                CheckBox {
                    id: ctrlCheckBoxLevel6
                    text: qsTr("6")
                }
                CheckBox {
                    id: ctrlCheckBoxLevel5
                    text: qsTr("5")
                }
            }

            Text {
                id: ctrlVoicesLabel
                x: 435 // orig 292
                y: 40
                width: 100
                text: "Voice:"
            }

            ComboBox {
                id: ctrlComboBoxVoice
                width: 55
                currentIndex: 1
                x: 435
                y: 60
                model: ListModel {
                    id: cbVoiceItems
                    ListElement { text: "1"; color: "Blue" }
                    ListElement { text: "2"; color: "Green" }
                    ListElement { text: "3"; color: "Brown" }
                    ListElement { text: "4"; color: "Purple" }
                }
            }

            Button {
                id: btnDeleteLevels
                x: 280
                y: 95
                width: 150
                height: 35
                text: qsTr("Delete")
                onClicked: {
                    if (selectLevels(false, true)) {
                        quit();
                    }
                }
            }

            Button {
                id: btnRevoiceLevels
                x: 280
                y: 55
                width: 150
                height: 35
                text: qsTr("Revoice")
                onClicked: {
                    if (selectLevels(true, false)) {
                        quit();
                    }
                }
            }

            Button {
                id: btnSelectLevels
                x: 280
                y: 15
                width: 150
                height: 35
                text: qsTr("Select Only")
                onClicked: {
                    if (selectLevels(false, false)) {
                        quit();
                    }
                }
            }


            Button {
                id: btnClose
                x: 50
                y: 170
                width: 125
                height: 35
                text: qsTr("Close")
                onClicked: {
                    console.log("Chord Levels closed.");
                    quit();
                }
            }

            Text {
                id: ctrlHintLabel
                x: 20
                y: 100
                width: 250
                text: qsTr("Hint: check the levels you want to select in order to delete, re-voice, or make a list-selection for further operations given a valid range selection.")
                font.italic: true
                color: "white"
                wrapMode: Text.WordWrap
                font.pointSize: 10
            }

        }

    }
}
